package name.dsouflis.aop.synthgetter;

import org.apache.bcel.Constants;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.*;

import java.util.Map;
import java.util.Set;

/** Replacement for the system ClassLoader that instruments some classes on load-time with output
 * statements that trace entering and exiting.
 * @author dsouflis
 * @version $Revision: 1.3 $
 */
public class SyntheticGetterClassLoader extends ClassLoader {

    String targetClassName;
    Map<String,String> idsByProperty;
    public final static String implSuffix = "$IMPL";
    private final String objectLocatorFieldName = "$objectLocator";
    private final String objectLocatorClassName = "name.dsouflis.aop.synthgetter.ObjectLocator";
    private final String objectLocatorGetMethodName = "getObject";
    private final String objectLocatorGetMethodSignature = "(Ljava/lang/String;)Ljava/lang/Object;";

    /** Creates new SyntheticGetterClassLoader
     * @param className The name of the class to instrument.&nbsp; All other classes will be loaded from parent.
     * @param idsByProperty A map of property names (whose getters will be synthesized) to ids to be passed to the
     * object locator
     */
    public SyntheticGetterClassLoader(String className, Map<String,String> idsByProperty) {
        this.targetClassName =className;
        this.idsByProperty =idsByProperty;
    }

    /** Loads the class, adding instrumentation if necessary.<br>
     * Delegates to the parent ClassLoader system classes.
     * @param name The name of the class to load.
     * @param resolve Whether the class should be resolved.
     * @throws ClassNotFoundException If the class cannot be found.
     * @return The loaded class.
     */
    protected Class loadClass(String name, boolean resolve) throws ClassNotFoundException {
        Class theClass=findLoadedClass(name);
        if(theClass!=null) {
            return theClass;
        }
        if(name.startsWith("java") || !name.equals(targetClassName+implSuffix)) {
            return getParent().loadClass(name);
        }
        theClass=findClass(name);
        if(theClass==null) {
            return null;
        }
        if(resolve) {
            resolveClass(theClass);
        }
        return theClass;
    }

    /** Defines the class.
     * @param className The name of the class.
     * @return The loaded class.
     */
    protected Class findClass(String className) {
        JavaClass clazz= classWithSyntheticGetters(className);
        if(clazz==null) {
            return null;
        }
        byte[] b = clazz.getBytes();
        return defineClass(className, b, 0, b.length);
    }

    /** Returns a JavaClass, instrumented if necessary.
     * @param className The class to modify.
     * @return The resulting JavaClass.
     */
    public JavaClass classWithSyntheticGetters(String className) {
        ClassGen classGen = new ClassGen(className, targetClassName, className + ".java", Constants.ACC_PUBLIC, new String[0]);
        JavaClass clazz = classGen.getJavaClass();
        clazz= synthesizeGetters(clazz);
        return clazz;
    }

    /** Adds synthesized getters to a JavaClass.
     * @param clazz The JavaClass.
     * @return The resulting JavaClass.
     */
    public JavaClass synthesizeGetters(JavaClass clazz) {
        JavaClass parentClass=null;
        try {
            parentClass = Repository.lookupClass(targetClassName);
        } catch (ClassNotFoundException e) {
            return null;
        }
        Method[] parentMethods = parentClass.getMethods();
        ClassGen cg=new ClassGen(clazz);
        addObjectLocator(cg);
        addConstructor(cg);
        Set<Map.Entry<String, String>> entries = idsByProperty.entrySet();
        for (Map.Entry<String, String> entry : entries) {
            String propertyName = entry.getKey();
            String beanId = entry.getValue();
            String getterName = new StringBuilder().append("get").append(Character.toUpperCase(propertyName.charAt(0))).append(propertyName.substring(1)).toString();
            int i = findGetter(parentMethods, getterName);
            Method getter = parentMethods[i];
            synthesizeGetter(getter, beanId, cg);
        }
        clazz=cg.getJavaClass();
        return clazz;
    }

/*
   Stack=1, Locals=1, Args_size=1
   0:   aload_0
   1:   invokespecial   #1; //Method name/dsouflis/aop/synthgetter/Foo."<init>":
()V
   4:   return
    */
    private void addConstructor(ClassGen cg) {
        ConstantPoolGen cp = cg.getConstantPool();
        int initMethodRef = cp.addMethodref(targetClassName, "<init>","()V");
        InstructionList il=new InstructionList();
        MethodGen newMethod=new MethodGen(
        Constants.ACC_PUBLIC,
        Type.VOID,
        Type.NO_ARGS,
        null,
        "<init>",
        cg.getClassName(),
        il,
        cp);

        il.append(new ALOAD(0));
        il.append(new INVOKESPECIAL(initMethodRef));
        il.append(new RETURN());

        newMethod.setMaxLocals();
        newMethod.setMaxStack();
        cg.addMethod(newMethod.getMethod());
        il.dispose();
    }

/*  Todo: add a proper setter for object locator, following the following template
   Stack=2, Locals=2, Args_size=2
   0:   aload_0
   1:   aload_1
   2:   putfield        #2; //Field $objectLocator:Lname/dsouflis/aop/synthgetter/ObjectLocator;
   5:   return
    */
    private void addObjectLocator(ClassGen cg) {
        cg.setAccessFlags(cg.getAccessFlags()&~Constants.ACC_ABSTRACT);
        FieldGen fieldGen = new FieldGen(Constants.ACC_PUBLIC|Constants.ACC_SYNTHETIC, Type.getType(new StringBuilder().append("L").append(objectLocatorClassName.replaceAll("\\.","/")).append(";").toString()), objectLocatorFieldName, cg.getConstantPool());
        cg.addField(fieldGen.getField());
    }

    private int findGetter(Method[] methods, String getterName) {
        for(int i=0; i<methods.length; i++) {
            Method method=methods[i];
            if(method.getName().equals(getterName)) {
                return i;
            }
        }
        return -1;
    }


    /** Synthesizes a getter, adding a body if it is abstract, or overriding it if it is not.
     * @param m The getter to synthesize (or override).
     * @param beanId The id to pass to the object locator in order to retrieve the object to return from the getter.
     * @param cg The ClassGen that represents the class under construction.
     */
    private void synthesizeGetter(Method m, String beanId, ClassGen cg) {
        ConstantPoolGen cp = cg.getConstantPool();
        int getObjectMethodRef = cp.addInterfaceMethodref(objectLocatorClassName, objectLocatorGetMethodName,
        objectLocatorGetMethodSignature);
        int beanIdRef = cp.addString(beanId);
        int objectLocatorFieldRef = cp.addFieldref(cg.getClassName(),objectLocatorFieldName, new StringBuilder().append("L").append(objectLocatorClassName.replaceAll("\\.", "/")).append(";").toString());

        int    flags = m.getAccessFlags();
        String name  = m.getName();
        String signature=m.getSignature();

        if(name.equals("<init>")) {
            return;
        }

        if(m.isAbstract()) {
            flags = flags&~Constants.ACC_ABSTRACT;
        }

        InstructionList il=new InstructionList();

        Type[] argTypes=Type.getArgumentTypes(signature);
        Type returnType=Type.getReturnType(signature);

        int returnTypeRef = cp.addClass(returnType.toString());

        MethodGen newMethod=new MethodGen(
        flags,
        returnType,
        argTypes,
        null,
        name,
        cg.getClassName(),
        il,
        cp);

        il.append(new ALOAD(0));
        il.append(new GETFIELD(objectLocatorFieldRef));
        il.append(new LDC(beanIdRef));
        il.append(new INVOKEINTERFACE(getObjectMethodRef, 2));
        il.append(new CHECKCAST(returnTypeRef));
        il.append(new ARETURN());

        newMethod.setMaxLocals();
        newMethod.setMaxStack();
        cg.addMethod(newMethod.getMethod());
        il.dispose();
    }
}
